#include "Parser.h"
#include <iostream>

Parser::Parser(SymbolTable* table, Symbolizer* symbolizer){
  this->symbol_table = table;
  this->symbolizer = symbolizer;
}

Expression* Parser::parseExpression(){
  // skip empty lines
  while (symbolizer->readNextSymbol()->isType(Symbol::EOL)) {}

  // Parse any function definitions
  while (parseDefinition()) {};

  // skip empty lines
  while (symbolizer->getSymbol()->isType(Symbol::EOL)) 
    symbolizer->readNextSymbol();

  // return if end of input
  if (symbolizer->getSymbol()->isType(Symbol::EOS))
    return NULL;
  
  Expression * exp = new Expression(symbolizer->getLineNumber());

  // Attempt to parse a binary expression
  ExpressionNode* node = parseBinaryExpression(exp);
  
  if(node == NULL)
    return NULL;

  exp->setRoot(node);
  
  Symbol* s = symbolizer->getSymbol();
  // the last symbol should be the end of the line/stream, or a '}'.
  if(!s->isType(Symbol::EOL) && 
     !s->isType(Symbol::EOS) &&
     !s->isType(Symbol::RBRACKET)){
    // A line can't start with a '*' or 'len'.
    if(s->isType(Symbol::UNARY) || s->isType(Symbol::BINARY))
      throw new ParseException(s->token, "primary expression preceding it", 32);
    else
      throw new ParseException(s->token, Symbol::eol.token);
  }
  return exp;
}

bool Parser::parseDefinition(){
  if(symbolizer->getSymbol()->isType(Symbol::DEF)){

    // parse the function type
    Symbol::SymbolType type = symbolizer->readNextSymbol()->type;
    if(type != Symbol::UNARY_DEF &&
       type != Symbol::BINARY_DEF &&
       type != Symbol::PREFIX_DEF)
      throw new ParseException(symbolizer->getSymbol()->token,
                               "unary, binary or prefix", 23);
    Symbol::SymbolType func_type;
    switch (type) {
    case Symbol::UNARY_DEF:
      func_type = Symbol::UNARY;
      break;
    case Symbol::BINARY_DEF:
      func_type = Symbol::BINARY;
      break;
    case Symbol::PREFIX_DEF:
      func_type = Symbol::PREFIX;
    default:
      break;
    }
    symbolizer->readNextSymbol();
      
    // parse the function name
    Symbol* namer = symbolizer->getSymbol();
    if(!namer->isType(Symbol::ID) &&
       !namer->isType(Symbol::UNARY) &&
       !namer->isType(Symbol::BINARY))
      throw new ParseException(symbolizer->getSymbol()->token,
                               "function name", 13);
      
    symbolizer->readNextSymbol();

    // create a new symbol table for the local scope
    SymbolTable* superscope = symbolizer->getSymbolTable();
    SymbolTable* subscope = new SymbolTable();
    subscope->setHardwired(superscope->getHardwired());
    symbolizer->setSymbolTable(subscope);

    // parse function arguments
    vector<Variable*>* args_vector;
    try {
      args_vector = parseFunctionArgs();
    } catch (ParseException e) {
      symbolizer->setSymbolTable(superscope);
      subscope->clear();
      delete subscope;
      delete namer;
      throw e;
    }
    
    Variable** args_array = new Variable*[args_vector->size()];
    for(unsigned int i=0; i < args_vector->size(); i++)
      args_array[i] = args_vector->at(i);

    // set the superscope for the local scope
    subscope->setSuperScope(superscope);

    // parse a code block
    vector<Expression*>* exp_vector = parseCodeBlock();
    
    Expression** exp_array = new Expression*[exp_vector->size()];
    for(unsigned int i=0; i < exp_vector->size(); i++)
      exp_array[i] = exp_vector->at(i);
    
    CustomFunction* val = new CustomFunction(func_type, namer->token.c_str(),
                                             exp_array, exp_vector->size(),
                                             args_array, args_vector->size());
    delete args_vector;
    delete exp_vector;
    
    // return to the global symbol table
    symbolizer->setSymbolTable(superscope);

    // clear and delete the local symbol table
    subscope->clear();
    delete subscope;

    // check if a code block was parsed
    if (val == NULL)
      throw new ParseException(symbolizer->getSymbol()->token, "{", 2);

    // remove any variable with the same name as the function from the symbol
    // table 
    superscope->remove(namer->token);
    // add the function
    superscope->set(val);
    cout << "New function: " << val->token << endl;

    return true;
  } else
    return false;
}


ExpressionNode* Parser::parseBinaryExpression(Expression* exp){
  // parse the left side of the expression
  ExpressionNode* left = parseUnaryExpression(exp);

  if(left == NULL)
    return NULL;

  // parse the binary operator
  ExpressionNode* root = NULL;
  while(symbolizer->getSymbol()->isType(Symbol::BINARY)){
    
    ExpressionNode* bin = exp->createNode(symbolizer->getSymbol(), 2);
    
    if (root == NULL) {
      // if this is the first operator then set left as the first branch
      exp->insert(left, bin, 0);
      root = bin;
    } else {
      // go down the right side of the tree, and put the operator below ones
      // that have lower precedence.
      ExpressionNode* node = root;
      while (node->getChildCount() == 2 &&
             symbol_table->hasHigherPrecedence(bin->symbol, node->symbol)){
	node = node->getChild(1);
      }

      if (node == root) {
        // bin becomes the new root
	exp->insert(node, bin, 0);
	root = bin;
      } else
        // insert bin between node and the node's parent
	exp->insert(bin, node->parent, 1);
    }
    
    symbolizer->readNextSymbol();

    // read a unary expresson to go to the right of bin
    ExpressionNode* right = parseUnaryExpression(exp);

    if(right == NULL)
      throw new ParseException(symbolizer->getSymbol()->token,
                               "primary or unary expression, following binary operator.", 56);
    else
      exp->insert(right, bin, 1);
  }
  
  if(root != NULL)
    return root;
  else
    return left;
}

ExpressionNode* Parser::parseUnaryExpression(Expression* exp){
  ExpressionNode* root = NULL;
  ExpressionNode* lastnode = NULL;

  // parse prefixes ('-' or '+')
  while (symbolizer->getSymbol()->isType(Symbol::PREFIX)) {
    ExpressionNode* node = exp->createNode(symbolizer->getSymbol(), 1);
    
    if (lastnode == NULL)
      root = lastnode = node;
    else
      lastnode = exp->insert(node, lastnode, 0);
    symbolizer->readNextSymbol();
  }
  
  // parse a primary expression
  ExpressionNode* primary = parsePrimary(exp);
  
  if (primary == NULL)
    return NULL;
  else if (root == NULL)
    root = primary;
  else
    exp->insert(primary, lastnode, 0);

  // parse the unary operator
  while (symbolizer->getSymbol()->isType(Symbol::UNARY)) {
    ExpressionNode* node = exp->createNode(symbolizer->getSymbol(), 1);
    exp->insert(root, node, 0);
    root = node;
    symbolizer->readNextSymbol();
  }

  return root;
}


ExpressionNode* Parser::parsePrimary(Expression* exp){
  ExpressionNode* node = NULL;

  // parse a constant or a variable
  if (symbolizer->getSymbol()->isType(Symbol::CONSTANT) || 
     symbolizer->getSymbol()->isType(Symbol::ID) ){
    
    node = exp->createNode(symbolizer->getSymbol(), 0);
    symbolizer->readNextSymbol();
  } else
    // or try to parse a sub expression 
    node = parseSubExpression(exp);
    
  return node;
}

ExpressionNode* Parser::parseSubExpression(Expression* exp){
  // read left parenthesis
  if(symbolizer->getSymbol()->isType(Symbol::LPAR)){
    
    ExpressionNode* ret = exp->createNode(symbolizer->getSymbol(), 1);
    symbolizer->readNextSymbol();

    // read the binary expression inside the parentheses
    ExpressionNode* node = parseBinaryExpression(exp);
    if(node == NULL)
      throw new ParseException(symbolizer->getSymbol()->token,
                               "binary expression", 18);
    else
      exp->insert(node, ret, 0);

    // read the right parenthesis
    if(symbolizer->getSymbol()->isType(Symbol::RPAR) == false)
      throw new ParseException(symbolizer->getSymbol()->token, ")", 2);
    else
      symbolizer->readNextSymbol();
    
    return  ret;
  }else
    return NULL;
}

vector<Expression*>* Parser::parseCodeBlock(){
  // read left bracket
  if (!symbolizer->getSymbol()->isType(Symbol::LBRACKET))
    throw new ParseException(symbolizer->getSymbol()->token, "{", 2);
  
  vector<Expression*>* exp_vector = new vector<Expression*>;

  // read expressions and store them in exp_vector
  while (!symbolizer->getSymbol()->isType(Symbol::RBRACKET)){
    Expression* exp = parseExpression();
    if(exp != NULL)
      exp_vector->push_back(exp);
  }
    
  symbolizer->readNextSymbol();
  return exp_vector;
}


vector<Variable*>* Parser::parseFunctionArgs () {
  // read left parenthesis
  if (!symbolizer->getSymbol()->isType(Symbol::LPAR))
    throw new ParseException(symbolizer->getSymbol()->token, "(", 2);
  
  vector<Variable*>* args_vector = new vector<Variable*>;

  // read variable names 
  while (!symbolizer->readNextSymbol()->isType(Symbol::RPAR)) {
    if (symbolizer->getSymbol()->isType(Symbol::ID)) 
      args_vector->push_back((Variable*)symbolizer->getSymbol());
    else {
      throw new ParseException(symbolizer->getSymbol()->token,
                               "identifier", 11);
    }
  }
  
  symbolizer->readNextSymbol();
  return args_vector;
}
