#ifndef __Parser_h__
#define __Parser_h__

#include "SymbolTable.h"
#include "Symbolizer.h"
#include "Symbol.h"
#include "Expression.h"
#include <exception>
#include <vector>

/*   Parser
   De Parser leest Symbols uit een Symbolizer en structureerd ze in een boomstructuur zoals
   die wordt ondersteund door Expression. 
   parseExpression() leest een hele expressie uit. Elke expressie wordt afgesloten met een 
   EOL of EOS symbool. Elk Symbol van het type BINARY(binaire operator) krijgt twee kinderen
   (vertakkingen), UNARY(unaire operatoren) en PREFIX('-' of '+') krijgen 1 kind. Variabelen
   en Constanten krijgen 0 kinderen toegewezen.
   De SymbolTable wordt geraadpleegt om uit te zoeken welke BINAIRE operator het meest 
   significant is.
   Een ParseException wordt gegeven als:
     - Voor een operator niet het juiste aantal kinderen uit de Symbolizer gelezen kan 
       worden.
     - Als twee pimaire operatoren elkaar volgen.
     - Als een unaire operator gegeven wordt zonder voorgaande primaire expressie.
     - Bij een subexpressie (beginnende met een '(') niet is afgesloten met een ')'.
*/

class ParseException: public exception{
 public:
  char* err;
	
  ParseException(string& found, string& expected){
    err = new char[23 + found.size() + expected.size()];
    sprintf(err, "Found %s when expecting %s", found.c_str(), expected.c_str());
  }
  ParseException(string& found, const char* expected, int len){
    err = new char[23 + found.size() + len];
    sprintf(err, "Found %s when expecting %s", found.c_str(), expected);
  }
	
  virtual const char* what() const throw(){
    return err;
  }
};

class Parser{
 private:
  SymbolTable* symbol_table;
  Symbolizer* symbolizer;
  
  //binary_expr = unary_expr { binary_op unary_expr }
  ExpressionNode* parseBinaryExpression(Expression* exp);	

  //unary_expr = {prefix} primary { unary_op }
  ExpressionNode* parseUnaryExpression(Expression* exp);
	
  //primary = prefix constant | '(' binary_expr ')' | ID
  ExpressionNode* parsePrimary(Expression* exp);
  ExpressionNode* parseSubExpression(Expression* exp);

  
  vector<Expression*>* parseCodeBlock();

  // function = 'def' ('binary' | 'unary' | 'prefix' ) ID func_args
  //       '{' binary_exp '}'
  bool parseDefinition();

  // func_args = '(' ID* ')'
  vector<Variable*>* parseFunctionArgs();
							
 public:
  Parser(SymbolTable* table, Symbolizer* symbolizer);
  Expression* parseExpression();
};

#endif
