#include "Symbol.h"
#include <iostream>

Symbol::Symbol(){}
Symbol::Symbol(SymbolType type, const char* token){
  this->type = type;
  this->token = string(token);
  pointercount = 0;
  // cout << "Symbol " << toString() << "(" << this <<") created" << endl;
}
Constant* Symbol::getConstant(){
  return new Scalar((double)0);
}
const char* Symbol::toString(){
  return token.c_str();
}
bool Symbol::isType(SymbolType type){
  if(this->type >= 8)
    return (this->type == type);
  else
    return ((this->type & type) == type);
}
Symbol* Symbol::getPointer(){
  pointercount++;
  return this;
}
void Symbol::removePointer(){
  pointercount--;
  if(pointercount == 0){
    //cout << "Symbol " << toString() << "(" << this << ") deleted" << endl;
    delete this;
  }
}

Symbol Symbol::eol = Symbol(Symbol::EOL, "end of line");
Symbol Symbol::eos = Symbol(Symbol::EOS, "end of stream");


Constant::Constant(const char* token):Symbol(CONSTANT, token){
}
Constant * Constant::getConstant(){
  return this;
}

Scalar::Scalar(const char* token): Constant(token){
  ctype = SCALAR;
  val = atof(token);
}
Scalar::Scalar(double d): Constant("Scalar"){
  token.clear();
  ostringstream s1;
  s1 << d;
  token.append(s1.str());
  ctype = SCALAR;
  val = d;
}
void Scalar::setValue(double value){
  val = value;
  ostringstream s1;
  s1 << val;
  token.clear();
  token.append(s1.str());
}
double Scalar::getValue(){
  return val;
}
Constant* Scalar::copy(){
  return new Scalar(val);
}

Vector::Vector(): Constant("Vector"){
  ctype = VECTOR;
  toString();
}
Vector::Vector(Symbol* s): Constant("Vector"){
  ctype = VECTOR;
  add(s);
  toString();
}
Vector::Vector(const char* token): Constant(token){
  ctype = VECTOR;
}
Vector::~Vector(){
  for(int i=0; i < size(); i++)
    get(i)->removePointer();
}

Symbol* Vector::add(Symbol * s){
  token.clear();
  list.push_back(s->getPointer());
  return s;
}
void Vector::insert(Symbol* s, int i){
  token.clear();
  list.insert(list.begin()+i, s->getPointer());
}
Symbol* Vector::get(int i){
  return list.at(i);
}
Symbol* Vector::remove(int i){
  token.clear();
  Symbol* s = get(i);
  s->removePointer();
  list.erase(list.begin()+i);
  return s;
}
int Vector::size(){
  return list.size();
}
Constant* Vector::copy(){
  Vector * v = new Vector();
  for(int i=0; i < size(); i++)
    v->add(get(i));
  return v;
}
const char* Vector::toString(){
  if(token.size() > 0)
    return token.c_str();
		
  if(size() == 0)
    token.append("empty");
  else{
    for(int i=0; i < size(); i++){
      token.append(get(i)->toString());
      if(get(i)->getConstant()->ctype == SCALAR)
	token.append(" ");
      else
	token.append("\n");
    }
  }
  return token.c_str();
}

Function::Function(SymbolType type, const char* token, Symbol * (*func)(Symbol * arg[], int argc)): Symbol(type, token){
  this->func = func;
}
Symbol * Function::call(Symbol* args[], int argc){
  return func(args, argc);
}

CustomFunction::CustomFunction(SymbolType type, const char* token, Expression* expression[], int expc, Variable* args[], int argc): Function(type, token, NULL){
  this->expression = expression;
  this->exp_count = expc;
  this->args = args;
  this->arg_count = argc;
  for(int i=0; i < arg_count; i++)
    args[i]->getPointer();
}
CustomFunction::~CustomFunction(){
  for(int i=0; i < exp_count; i++)
    delete expression[i];
  for(int i=0; i < arg_count; i++)
    args[i]->removePointer();
}

Symbol * CustomFunction::call(Symbol* args[], int argc){

  if (this->arg_count > 0)
    this->args[0]->setValue(args[0]->getConstant());

  if (this->type == Symbol::BINARY && this->arg_count > 1)
    this->args[1]->setValue(args[1]->getConstant());
  
  Executrix* executrix = new Executrix();
  Symbol* returnval = NULL;
  
  for (int i = 0; i < exp_count; i++) {
    Expression* clone = expression[i]->clone();
    Symbol* s = executrix->execute(clone);

    if (i == exp_count - 1)
      returnval = s->getConstant()->copy();
    delete clone;
  }

  if(this->arg_count > 0)
     this->args[0]->setValue(NULL);
  
  delete executrix;
  return returnval;
}

Variable::Variable(const char* token):Symbol(ID, token){
  value = NULL;
}
Variable::~Variable(){
  if(value != NULL)
    value->removePointer();
}

Constant * Variable::getConstant(){
  if(value == NULL)
    value = new Scalar(0.0);
	
  return value;
}
void Variable::setValue(Constant * val){
  if(value != NULL)
    value->removePointer();
  value = val;
  if(value != NULL)
    value->getPointer();
}
